/*
 * Validator.java
 *
 * ECHA is providing program source code to facilitate the preparation of
 * pre-registration XML files. The code shall serve as example and
 * registrants are under no circumstances obliged to make use of this service
 * for submitting data to ECHA. In no event shall ECHA be liable for any
 * damages, direct, incidental or consequential (including through negligence
 * or otherwise) which may result from the use of the data provided, even if
 * ECHA has been advised of the possibility of such damages.
 * The users must assume the entire risk of using this source code.
*/

package eu.europa.echa.reach.preregistration.examples.preregcheck;

import java.awt.Color;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import javax.xml.XMLConstants;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;
import javax.xml.transform.Source;
import javax.xml.transform.sax.SAXSource;
import javax.xml.validation.Schema;
import javax.xml.validation.SchemaFactory;
import org.apache.log4j.Logger;
import org.w3c.dom.DOMImplementation;
import org.w3c.dom.bootstrap.DOMImplementationRegistry;
import org.w3c.dom.ls.DOMImplementationLS;
import org.w3c.dom.ls.LSInput;
import org.w3c.dom.ls.LSResourceResolver;
import org.xml.sax.Attributes;
import org.xml.sax.InputSource;
import org.xml.sax.Locator;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import org.xml.sax.helpers.DefaultHandler;

/**
 *
 * @author Csk Gbor
 */
public class Validator {
    
    public interface MessageWriter {
        void write(final String message, final Color color);
    }
    
    private static final Logger logger = Logger.getLogger(Validator.class);
    
    private SAXParser parser;
    private Schema schema;
    private MessageWriter messageWriter;
    private int ErrorCount;
    private int WarningCount;
    
    public Validator() {
    }
    
    public MessageWriter getMessageWriter() {
        return messageWriter;
    }
    public void setMessageWriter(MessageWriter messageWriter) {
        this.messageWriter = messageWriter;
    }
    
    public void validate(final String[] fileAndDirNames) {
        SAXParser parser = getParser();
        
        File[] files = new File[fileAndDirNames.length];
        for (int i=0; i<fileAndDirNames.length; i++) {
            files[i] = new File(fileAndDirNames[i]);
        }
        validate(files);
    }
    
    private void validate(final File[] files) {
        if (files == null) {
            return;
        }
        for (File file: files) {
            if (file.isDirectory()) {
                validate(file.listFiles());
            } else {
                try {
                    SAXParser parser = getParser();
                    messageWriter.write("Checking file: " + file.getAbsolutePath(), Color.BLACK);
                    ErrorCount = WarningCount = 0;
                    parser.parse(file, new ParserEventHandler());
                    if (ErrorCount == 0 && WarningCount == 0) {
                        messageWriter.write("...OK", Color.GREEN);
                    } else {
                        messageWriter.write("\nErrors: " + ErrorCount + "  Warnings: " + WarningCount , Color.BLACK);
                    }
                } catch (IOException ex) {
                    logger.error(ex);
                    messageWriter.write("\n", null);
                    messageWriter.write(ex.getMessage(), Color.RED);
                } catch (SAXException ex) {
                    logger.error(ex);
                    messageWriter.write("\n", null);
                    messageWriter.write(ex.getMessage(), Color.RED);
                }
                messageWriter.write("\n", Color.BLACK);
            }
        }
    }
    
    private SAXParser getParser() {
        if (parser == null) {
            try {
                SAXParserFactory spf = SAXParserFactory.newInstance();
                spf.setNamespaceAware(true);
                //WARNING: do not set spf.setValidating(true);
                //it turns on validation against DTD, but we need XSD
                spf.setValidating(false);
                spf.setSchema(getSchema());
                parser = spf.newSAXParser();
            } catch (SAXException ex) {
                logger.error(ex);
            } catch (ParserConfigurationException ex) {
                logger.error(ex);
            }
        }
        return parser;
    }
    
    private Schema getSchema() throws SAXException {
        if (schema == null) {
            schema = loadSchema();
        }
        return schema;
    }
    
    private Schema loadSchema() throws SAXException {
        // create a SchemaFactory that conforms to W3C XML Schema
        SchemaFactory schemaFactory = SchemaFactory.newInstance(XMLConstants.W3C_XML_SCHEMA_NS_URI);
        
        // set your error handler to catch errors during schema construction
        //schemaFactory.setErrorHandler(new SchemaErrorHandler());
        
        //
        schemaFactory.setResourceResolver(new LSResourceResolver() {
            public LSInput resolveResource(String type, String namespaceURI, String publicId, String systemId, String baseURI) {                DOMImplementation dom = null;
                try {
                    DOMImplementationRegistry registry = DOMImplementationRegistry.newInstance();
                    dom = registry.getDOMImplementation("XML 1.0 LS 3.0");
                } catch (Exception ex) {
                    logger.error(ex);
                }
                
                LSInput lsInput = null;
                if (dom != null) {
                    String resourceFileName = null;
                    DOMImplementationLS domLS = (DOMImplementationLS) dom;
                    lsInput = domLS.createLSInput();
                    if (systemId.equals("http://www.w3.org/2001/xml.xsd")) {
                        resourceFileName = "xml.xsd";
                    } else {
                        resourceFileName = systemId;
                    }
                    logger.info("Loading XML resource: " + namespaceURI + ": " + resourceFileName);
                    InputStream is = ClassLoader.getSystemResourceAsStream(resourceFileName);
                    assert is != null: "Schema cannot be loaded systemId: " + systemId;
                    lsInput.setByteStream(is);
                }
                
                return lsInput;
            }
        });
        
        //InputStream is = new FileInputStream("schemas" + File.separator + "PreRegistration.xsd");
        InputStream is = ClassLoader.getSystemResourceAsStream("PreRegistration.xsd");
        assert is != null: "Schema cannot be loaded";
        Source source = new SAXSource(new InputSource(is));
        Schema schema = schemaFactory.newSchema(source);
        return schema;
    }
    
    private class ParserEventHandler extends DefaultHandler {
        private Locator locator;
        private boolean rootElement;
        
        public ParserEventHandler() {
        }
        
        public void error(SAXParseException e) throws SAXException {
            ErrorCount++;
            messageWriter.write("\n" + e.getMessage(), Color.RED);
            messageWriter.write("\n" + e.getSystemId(), Color.BLUE);
            messageWriter.write("\nline: " + e.getLineNumber(), Color.BLUE);
            messageWriter.write(", column: " + e.getColumnNumber(), Color.BLUE);
            super.error(e);
        }
        
        public void warning(SAXParseException e) throws SAXException {
            WarningCount++;
            messageWriter.write("\n" + e.getMessage(), Color.ORANGE);
            messageWriter.write("\n" + e.getSystemId(), Color.BLUE);
            messageWriter.write("\nline: " + e.getLineNumber(), Color.BLUE);
            messageWriter.write(", column: " + e.getColumnNumber(), Color.BLUE);
            super.warning(e);
        }
        
        public void startElement(String uri, String localName, String qName, Attributes attributes) throws SAXException {
            if (rootElement) {
                rootElement = false;
                if (uri.equals("http://echa.europa.eu/schemas/preRegistration")) {
                    if (localName != "SimplePreRegistration" && localName != "BulkPreRegistration") {
                        warning(new SAXParseException("'" + qName + "' is invalid root element. Only 'SimplePreRegistration' and 'BulkPreRegistration' are valid root elements.", locator));
                    }
                } else {
                    error(new SAXParseException("The root namespace must be 'http://echa.europa.eu/schemas/preRegistration'", locator));
                }
                
                super.startElement(uri, localName, qName, attributes);
            }
        }
        
        public void setDocumentLocator(Locator locator) {
            this.locator = locator;
            super.setDocumentLocator(locator);
        }
        
        public void startDocument() throws SAXException {
            rootElement = true;
            super.startDocument();
        }
    }
}
